/*=============================================================================
Turing Machine Debugger
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2004
  -- License restrictions apply (Dual FAL/GPL2)
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <sys/select.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX.h>
#include "wx/wx.h"
#include "ops.h"
#include "mach.hpp"
#include "sim.hpp"

static size_t count_trailing_zeroes(const char *);

BEGIN_EVENT_TABLE(Simulator, wxFrame)
    EVT_BUTTON(SOP_INIT, Simulator::cb_init)
    EVT_BUTTON(SOP_STEP, Simulator::cb_step)
    EVT_BUTTON(SOP_RUN,  Simulator::cb_run)
    EVT_TIMER(SOP_RUN_TIMER, Simulator::do_run)
END_EVENT_TABLE()

//-----------------------------------------------------------------------------
Simulator::Simulator(const char *title, const wxPoint &pos,
 const wxSize &size) :
    wxFrame(NULL, wxID_ANY, title, pos, size)
{
    // Graphical interface instantiation
    wxBoxSizer *top = new wxBoxSizer(wxVERTICAL);

    tape = NULL;
    mach = NULL;

    { // Tape
        wxStaticBoxSizer *packer;
        wxStaticBox *sbox;
        wxFont tt(14, wxTELETYPE, wxNORMAL, wxBOLD, false, "courier");

        sbox   = new wxStaticBox(this, -1, "Tape (tape.cpp)");
        packer = new wxStaticBoxSizer(sbox, wxVERTICAL);
        t_data = new wxStaticText(this, -1, "(Tape uninitialized)");
        t_head = new wxStaticText(this, -1, "(Head uninitialized)");
        t_data->SetFont(tt);
        t_head->SetFont(tt);
        packer->Add(t_data, 1, wxALL | wxALIGN_LEFT | wxGROW, 3);
        packer->Add(t_head, 1, wxALL | wxALIGN_LEFT | wxGROW, 3);
        top->Add(packer, 0, wxALL | wxALIGN_LEFT | wxGROW, 3);
    }

    { // Machine
        wxStaticBoxSizer *packer;
        wxStaticText *l_ms;
        wxStaticBox *sbox;
        wxBoxSizer *ctlpack;
        wxButton *b_step, *b_until;

        sbox    = new wxStaticBox(this, -1, "Turing Machine (mach.cpp)");
        packer  = new wxStaticBoxSizer(sbox, wxVERTICAL);
        m_state = new wxTextCtrl(this, -1, "", wxDefaultPosition,
                   wxDefaultSize, wxTE_MULTILINE);

        ctlpack = new wxBoxSizer(wxHORIZONTAL);
        m_until = new wxTextCtrl(this, -1, "", wxDefaultPosition,
                   wxDefaultSize, wxTE_RIGHT);
        b_until = new wxButton(this, SOP_RUN, "Step until");
        b_step  = new wxButton(this, SOP_STEP, "Single step");
        m_hex   = new wxCheckBox(this, -1, "Hex state");
        m_itv   = new wxTextCtrl(this, -1, "250");
        l_ms    = new wxStaticText(this, -1, "ms");

        ctlpack->Add(m_until, 0, wxALL | wxALIGN_BOTTOM, 3);
        ctlpack->Add(b_until, 0, wxALL | wxALIGN_BOTTOM, 3);
        ctlpack->Add(b_step, 0, wxALL | wxALIGN_BOTTOM, 3);
        ctlpack->Add(m_hex, 0, wxALL | wxALIGN_BOTTOM, 3);
        ctlpack->Add(m_itv, 0, wxALL | wxALIGN_BOTTOM, 3);
        ctlpack->Add(l_ms, 0, wxALL | wxALIGN_BOTTOM, 3);
        packer->Add(m_state, 1, wxALL | wxALIGN_LEFT | wxGROW, 3);
        packer->Add(ctlpack, 0, wxALL | wxALIGN_RIGHT, 3);
        top->Add(packer, 1, wxALL | wxALIGN_LEFT | wxGROW, 3);
    }

    { // Initialiser
        wxStaticBoxSizer *packer;
        wxStaticBox *sbox;
        wxButton *button;
        wxChoice *machtype;
        const char **md_names = (const char **)malloc(sizeof(const char *) * MACHTYPES_NUM);

        for(size_t i = 0; i < MACHTYPES_NUM; ++i) {
            md_names[i] = Machtypes[i].name;
        }

        sbox   = new wxStaticBox(this, -1, "Initializer (sim.cpp)");
        packer = new wxStaticBoxSizer(sbox, wxHORIZONTAL);
        i_data = new wxTextCtrl(this, -1, "00111100000000");
        i_mt   = new wxChoice(this, -1, wxDefaultPosition, wxDefaultSize,
                  MACHTYPES_NUM, (const wxString *)md_names);
        button = new wxButton(this, SOP_INIT, "Update");

        packer->Add(i_data, 1, wxALL | wxALIGN_BOTTOM, 3);
        packer->Add(i_mt, 0, wxALL | wxALIGN_BOTTOM, 3);
        packer->Add(button, 0, wxALL | wxALIGN_BOTTOM, 3);
        top->Add(packer, 0, wxALL | wxALIGN_LEFT | wxGROW, 3);
        free(md_names);
    }

    SetSizer(top);
    return;
}

void Simulator::cb_init(wxCommandEvent &ev) {
    /* Take the input string in the "Initialiser" section and selected machine
    type and initialize the new tape/machine, respectively. */
    if(mach != NULL) { delete mach; }
    if(tape != NULL) { delete tape; }

    { // Initialize machine
        size_t mtnum;
        if((mtnum = i_mt->GetSelection()) == -1) {
            m_state->AppendText("No machine type selected\n");
            return;
        }

        mach = new Machine(Machtypes[mtnum].ops);
    }

    { // Initialize tape
        wxString dt = i_data->GetValue();
        size_t scur = strlen(dt), szero = count_trailing_zeroes(dt);
        char *ptr, *dv;

        if(szero < 2 * scur + 5) {
            dv = (char *)malloc(2 * scur + 6);
            memset(dv, '0', 2 * scur + 5);
            dv[2 * scur + 5] = '\0';
            memcpy(dv, dt, scur);
        } else {
            dv = HX_strdup(dt);
        }

        if((ptr = strrchr(dv, '1')) == NULL) {
            t_data->SetLabel("No '1' found in string");
            t_head->SetLabel("(Head uninitialized)");
            return;
        }

        tape = new Tape(dv, ptr - dv);
        free(dv);
    }

    update_tape();
    m_state->Clear();
    m_state->AppendText("Turing Machine initialized\n");
    show_next_step();
    return;
}

void Simulator::cb_run(wxCommandEvent &ev) {
    // Callback associated with the "Step until" button
    wxString v = m_until->GetValue();
    int i = strtol(m_itv->GetValue(), NULL, 0);
    if(i == 0) { i = 100; }

    if(!v) {
        m_stop = -1;
    } else {
        m_stop = strtol(v, NULL, 0);
    }
    m_timer = new wxTimer(this, SOP_RUN_TIMER);
    m_timer->Start(i);
    return;
}

void Simulator::do_run(wxTimerEvent &ev) {
    int m = do_step();
    if(m == 'E' /* sig */ || m == 'H' /* finish */ ||
     mach->state == m_stop /* breakpoint */) {
        m_timer->Stop();
        return;
    }

    return;
}

void Simulator::cb_step(wxCommandEvent &ev) {
    /* Need to get rid of the "ev" parameter, since it is a non-pointer and
    do_step() is called elsewhere too. */
    do_step();
    return;
}

int Simulator::do_step(void) {
    char r, w, m;

    if(tape == NULL || mach == NULL) {
        m_state->AppendText("Tape is not initialized!\n");
        return 'E';
    }

    r = tape->read();           // Read byte
    w = mach->out(r);
    tape->write(w);             // Write byte

    switch(m = mach->mov(r)) {  // Execute movement
        case 'E':
            m_state->AppendText("Error!\n");
            break;
        case 'L':
            tape->movl();
            break;
        case 'R':
            tape->movr();
            break;
        case 'H': {
            char buf[128];
            snprintf(buf, sizeof(buf), "Finished! (%d movs)\n", mach->iic);
            m_state->AppendText(buf);
            break;
        }
        case '-':
            break;
        default:
            m_state->AppendText("Undefined MOV instruction!\n");
            break;
    }

    // Let it change to the next status
    mach->tick(r);

    // Update GUI representation
    update_tape();
    if(m != 'E' && m != 'H') { show_next_step(); }
    return m;
}

void Simulator::show_next_step(void) {
    /* Display the instruction and action the machine would execute next.
    Useful for single step debugging. */
    char buf[128], r = tape->read();

    if(m_hex->IsChecked()) {
        snprintf(buf, sizeof(buf),
         "state=0x%02X, read='%c', write='%c', mov='%c', newstate=0x%02X\n",
         mach->state, r, mach->out(r), mach->mov(r), mach->next(r));
    } else {
        snprintf(buf, sizeof(buf),
         "state=%d, read='%c', write='%c', mov='%c', newstate=%d\n",
         mach->state, r, mach->out(r), mach->mov(r), mach->next(r));
    }

    m_state->AppendText(buf);
    return;
}

void Simulator::update_tape(void) {
    // Update the GUI tape with the real tape contents and the machine's head
    size_t len = strlen(tape->data);
    char *buf = (char *)malloc(len + 1);

    memset(buf, ' ', len);
    buf[len] = '\0';
    buf[tape->pos] = '^';

    t_data->SetLabel(tape->data);
    t_head->SetLabel(buf);
    free(buf);
    return;
}

//-----------------------------------------------------------------------------
static size_t count_trailing_zeroes(const char *str) {
    const char *ptr = str + strlen(str) - 1;
    size_t i = 0;

    if(ptr == NULL) { return 0; }
    while(ptr >= str && *ptr == '0') {
        ++i;
        --ptr;
    }

    return i;
}

//==[ End of file ]============================================================
